@@ -20,7 +20,7 @@
 #
 #
 import logging
-from typing import TYPE_CHECKING, List, Optional, Tuple, Union, cast
+from typing import TYPE_CHECKING, cast
 
 import attr
 
@@ -99,14 +99,14 @@ class UserPaginateResponse:
     """This is very similar to UserInfo, but not quite the same."""
 
     name: str
-    user_type: Optional[str]
+    user_type: str | None
     is_guest: bool
     admin: bool
     deactivated: bool
     shadow_banned: bool
-    displayname: Optional[str]
-    avatar_url: Optional[str]
-    creation_ts: Optional[int]
+    displayname: str | None
+    avatar_url: str | None
+    creation_ts: int | None
     approved: bool
     erased: bool
     last_seen_ts: int
@@ -180,17 +180,17 @@ class DataStore(
         self,
         start: int,
         limit: int,
-        user_id: Optional[str] = None,
-        name: Optional[str] = None,
+        user_id: str | None = None,
+        name: str | None = None,
         guests: bool = True,
-        deactivated: Optional[bool] = None,
-        admins: Optional[bool] = None,
+        deactivated: bool | None = None,
+        admins: bool | None = None,
         order_by: str = UserSortOrder.NAME.value,
         direction: Direction = Direction.FORWARDS,
         approved: bool = True,
-        not_user_types: Optional[List[str]] = None,
+        not_user_types: list[str] | None = None,
         locked: bool = False,
-    ) -> Tuple[List[UserPaginateResponse], int]:
+    ) -> tuple[list[UserPaginateResponse], int]:
         """Function to retrieve a paginated list of users from
         users list. This will return a json list of users and the
         total number of users matching the filter criteria.
@@ -216,7 +216,7 @@ class DataStore(
 
         def get_users_paginate_txn(
             txn: LoggingTransaction,
-        ) -> Tuple[List[UserPaginateResponse], int]:
+        ) -> tuple[list[UserPaginateResponse], int]:
             filters = []
             args: list = []
 
@@ -311,7 +311,7 @@ class DataStore(
                 """
             sql = "SELECT COUNT(*) as total_users " + sql_base
             txn.execute(sql, args)
-            count = cast(Tuple[int], txn.fetchone())[0]
+            count = cast(tuple[int], txn.fetchone())[0]
 
             sql = f"""
                 SELECT name, user_type, is_guest, admin, deactivated, shadow_banned,
@@ -351,9 +351,7 @@ class DataStore(
 
     async def search_users(
         self, term: str
-    ) -> List[
-        Tuple[str, Optional[str], Union[int, bool], Union[int, bool], Optional[str]]
-    ]:
+    ) -> list[tuple[str, str | None, int | bool, int | bool, str | None]]:
         """Function to search users list for one or more users with
         the matched term.
 
@@ -366,9 +364,7 @@ class DataStore(
 
         def search_users(
             txn: LoggingTransaction,
-        ) -> List[
-            Tuple[str, Optional[str], Union[int, bool], Union[int, bool], Optional[str]]
-        ]:
+        ) -> list[tuple[str, str | None, int | bool, int | bool, str | None]]:
             search_term = "%%" + term + "%%"
 
             sql = """
@@ -379,13 +375,13 @@ class DataStore(
             txn.execute(sql, (search_term,))
 
             return cast(
-                List[
-                    Tuple[
+                list[
+                    tuple[
                         str,
-                        Optional[str],
-                        Union[int, bool],
-                        Union[int, bool],
-                        Optional[str],
+                        str | None,
+                        int | bool,
+                        int | bool,
+                        str | None,
                     ]
                 ],
                 txn.fetchall(),
