@@ -24,16 +24,12 @@ from itertools import islice
 from typing import (
     Callable,
     Collection,
-    Dict,
     Generator,
     Iterable,
     Iterator,
-    List,
     Mapping,
     Protocol,
-    Set,
     Sized,
-    Tuple,
     TypeVar,
 )
 
@@ -52,7 +48,7 @@ class _SelfSlice(Sized, Protocol):
     def __getitem__(self: S, i: slice) -> S: ...
 
 
-def batch_iter(iterable: Iterable[T], size: int) -> Iterator[Tuple[T, ...]]:
+def batch_iter(iterable: Iterable[T], size: int) -> Iterator[tuple[T, ...]]:
     """batch an iterable up into tuples with a maximum size
 
     Args:
@@ -80,7 +76,7 @@ def chunk_seq(iseq: S, maxlen: int) -> Iterator[S]:
 
 def partition(
     iterable: Iterable[T], predicate: Callable[[T], bool]
-) -> Tuple[List[T], List[T]]:
+) -> tuple[list[T], list[T]]:
     """
     Separate a given iterable into two lists based on the result of a predicate function.
 
@@ -114,8 +110,8 @@ def sorted_topologically(
 
     # This is implemented by Kahn's algorithm.
 
-    degree_map = {node: 0 for node in nodes}
-    reverse_graph: Dict[T, Set[T]] = {}
+    degree_map = dict.fromkeys(nodes, 0)
+    reverse_graph: dict[T, set[T]] = {}
 
     for node, edges in graph.items():
         if node not in degree_map:
@@ -164,8 +160,8 @@ def sorted_topologically_batched(
     persisted.
     """
 
-    degree_map = {node: 0 for node in nodes}
-    reverse_graph: Dict[T, Set[T]] = {}
+    degree_map = dict.fromkeys(nodes, 0)
+    reverse_graph: dict[T, set[T]] = {}
 
     for node, edges in graph.items():
         if node not in degree_map:
